/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: region.cpp 604 2018-12-06 14:50:52Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C)  Niv Drory <drory@mpe.mpg.de>
 *                         Claus A. Goessl <cag@usm.uni-muenchen.de>
 *                         Arno Riffeser <arri@usm.uni-muenchen.de>
 *                         Jan Snigula <snigula@usm.uni-muenchen.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */


#include <ltl/util/region.h>

namespace util {

// implementation
   Region::Region(const std::size_t N) :
      N_(N), start_(new int [2*N_]), end_(start_ + N_)
   {
      for(std::size_t i = 0; i < N_; ++i)
      {
	 start_[i] = 0;
	 end_[i] = 0;
      }
   }

   Region::Region(const std::string& s, const std::size_t N) :
      N_(N), start_(new int [2*N_]), end_(start_ + N_)
   {
      // evaluate N_ from string does not comply with int *const start_ and int *const end_
      // you may use next few lines of code to evaluate No dimonsions from string
      // if( N_ == 0 )
      // {
      //    ++N_;
      //    fostring::const_iterator s_i = s.begin; s_i != s.end(); ++s_i)
      //       if(*s_i == ',') ++N_;
      // }

      if( (s[0] != '[') || (s[s.length()-1] != ']') )
	 throw(UException( std::string("Region string ") + s + std::string(" not within [ ].") ));

      std::string str = s.substr( 1, s.length()-2 );
      std::string token;
      std::string::size_type end = 1;

      // comma separated string of dim_i_start:dim_i_end
      for( std::size_t i = 1; i <= N_; ++i )
      {
	 end = str.find_first_of( ":", 0 );
	 if( end != str.npos )
	 {
	    token = str.substr( 0, end );
	    str.erase( 0, end+1 );
	 }
	 else
	    throw(UException( std::string("Missing startvalue endvalue separator ':' in ") + s ));

	 const int startvalue = atoi( token.c_str() );
	 end = str.find_first_of( ",", 0 );
	 if( end != str.npos )
	 {
	    token = str.substr( 0, end );
	    str.erase( 0, end+1 );
	 }
	 else
	 {
	    if(i < N_)
	       throw(UException( std::string("Missing axes separator ',' in ") + s ));
	    token = str.substr( 0, str.length() );
	    str.erase(0, str.length() );
	 }
	 const int endvalue = atoi( token.c_str() );
	 setRange(i,
		  ( (endvalue > startvalue) ? startvalue : endvalue ),
		  ( (endvalue > startvalue) ? endvalue : startvalue )); // swap if s>e
      }
   }

   Region::Region(const Region& other) :
      N_(other.N_), start_(new int [2*N_]), end_(start_ + N_)
   {
      setRegion(other);
   }

   int Region::getStart(const std::size_t i) const
   {
      if( (i < 1) || (i > N_) )
	 throw UException(error_get_dim);
      return start_[i-1];
   }

   int Region::getEnd(const std::size_t i) const
   {
      if( (i < 1) || (i > N_) )
	 throw UException(error_get_dim);
      return end_[i-1];
   }

   std::size_t Region::getLength() const
   {
      std::size_t totlength = 1;
      for(std::size_t i = 0; i < N_; ++i)
	 totlength *= abs(end_[i] - start_[i]) + 1;
      return totlength;
   }

   std::size_t Region::getLength(const std::size_t i) const
   {
      if( (i < 1) || (i > N_) )
	 throw UException(error_get_dim);
      return abs(end_[i-1] - start_[i-1]) + 1;
   }

   Region Region::getSlice(const std::size_t startdim, const std::size_t enddim) const
   {
      if( (startdim > enddim) || (startdim < 1) || (enddim > N_))
	 throw UException(error_get_dim);
      Region slice( std::size_t(1 + enddim - startdim) );
      for(std::size_t i = 0, j = startdim - 1; j < enddim; ++i, ++j)
      {
	 slice.start_[i] = start_[j];
	 slice.end_[i] = end_[j];
      }
      return slice;
   }

   void Region::setStart(const std::size_t i, const int s)
   {
      if( (i < 1) || (i > N_) )
	 throw UException(error_set_dim);
      start_[i-1] = s;
   }

   void Region::setEnd(const std::size_t i, const int e)
   {
      if( (i < 1) || (i > N_) )
	 throw UException(error_set_dim);
      end_[i-1] = e;
   }

   void Region::setRange(const std::size_t i, const int s, const int e)
   {
      if( (i < 1) || (i > N_) )
	 throw UException(error_set_dim);
      start_[i-1] = s;
      end_[i-1] = e;
   }

   void Region::setRegion(const Region& other)
   {
      if( N_ != other.N_ )
	 throw UException(error_set_dim);
      for(std::size_t i = 0; i < N_; ++i)
      {
	 start_[i] = other.start_[i];
	 end_[i] = other.end_[i];
      }
   }

   std::string Region::toString() const
   {
      std::string s = "[";
      std::size_t i;
      char buf[255];
      for( i = 0; i < (N_-1); ++i )
      {
	 snprintf( buf, sizeof(buf), "%d:%d,", start_[i], end_[i] );
	 s += std::string(buf);
      }
      snprintf( buf, sizeof(buf), "%d:%d", start_[i], end_[i] );
      s += std::string(buf) + std::string("]");
      return s;
   }

   const std::string Region::error_get_dim =
      "request for region parameter of illegal dimension";
   const std::string Region::error_set_dim = 
      "request to set region parameter of illegal dimension";

}
