/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: stringfac.h 599 2018-08-22 16:13:53Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C)  Niv Drory <drory@mpe.mpg.de>
 *                         Claus A. Goessl <cag@usm.uni-muenchen.de>
 *                         Arno Riffeser <arri@usm.uni-muenchen.de>
 *                         Jan Snigula <snigula@usm.uni-muenchen.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

#ifndef __STRINGFAC_H__
#define __STRINGFAC_H__

#include <ltl/config.h>

#include <string>
#include <iostream>
#include <iomanip>

#ifdef HAVE_SSTREAM
#ifdef HAVE_IOS

#define UTIL_USING_SSTREAM
#include <sstream>

#endif
#endif

#ifndef UTIL_USING_SSTREAM

#ifdef HAVE_STRSTREAM
#include <strstream>

#else
#error <sstream> or <strstream> needed!
#endif

#endif

#include <cstdlib>
#include <cmath>
#include <climits>
#include <cfloat>

#include <ltl/util/u_exception.h>

namespace util {

//! This class returns a number in the requested type.
/*! In case of string output this is a right adjusted string
  of selected length and (if possible) precision.
  The string is filled with leading blanks to match width.
*/
   class StringFactory
   {
   protected:
      static const std::string factory_text;
      static const std::string error_text;

      /*!
       * Convert float/double to string
       * \param format: char to select the _floatfield_ format flag
       *   - 'd': default floating-point notation
       *   - 'f': fixed-point notation
       *   - 'e': scientific notation
       */
      template<class T>
      inline static std::string floatToString(const T xd, const int width,
					      const int precision, const char format);
      template<class T>
      inline static std::string intToString(const T xd, const int width);
      
      /*!
       * set the _floatfield_ format flag
       * \param s: `std::ostringstream` where to write
       * \param format: char to select the _floatfield_ format flag
       *   - 'd': default floating-point notation
       *   - 'f': fixed-point notation
       *   - 'e': scientific notation
       */
      void static changeFormat(std::ostringstream& s, const char format);

   public:
      /*!
       * Convert number to string
       * \param format: (for double and float only) char to select the _floatfield_ format flag
       *   - 'd': default floating-point notation
       *   - 'f': fixed-point notation
       *   - 'e': scientific notation
       */
      static std::string toString(const double xd,
				  const int width = DBL_DIG + 7,
				  const int precision = DBL_DIG,
				  const char format = 'd');
      static std::string toString(const float xd,
				  const int width = FLT_DIG + 7,
				  const int precision = FLT_DIG,
				  const char format = 'd');
      static std::string toString(const long xd,
				  const int width = ( (sizeof(long) * 8) / 3) + 1);
      static std::string toString(const int xd,
				  const int width = ( (sizeof(int) * 8) / 3) + 1);

      //! Return \e xd in dd:mm:ss[.sss] format.
      /*! Precision gives No second decimals. */
      static std::string toMinSecString(const double xd,
					const int precision = 0);

      //! Convert \e xs to integer number.
      static int toInt(const std::string& xs)
      {
         return atoi(xs.c_str());
      }
      //! Convert \e xs to long integer number.
      static long toLong(const std::string& xs)
      {
         return atol(xs.c_str());
      }
      //! Convert \e xs to float number.
      /*! dd:mm:ss[.sss] strings are automatically recognized. */
      static float toFloat(const std::string& xs)
      {
         return float(toDouble(xs));
      }
      //! Convert \e xs to double number.
      /*! dd:mm:ss[.sss] strings are automatically recognized. */
      static double toDouble(std::string xs);

   };

#ifdef UTIL_USING_SSTREAM
   template<class T>
   std::string StringFactory::floatToString(const T xd,
					    const int width, const int precision,
					    const char format)
   {
      std::string s_string;
      {
	 std::ostringstream s_stream;
	 changeFormat(s_stream, format);
	 s_stream << std::setiosflags(std::ios::right | std::ios::adjustfield)
		  << std::setw(width) << std::setprecision(precision) << xd;
	 s_string = s_stream.str();
//       std::cerr << xd << ", width " << width << ", prec " << precision 
//            << ", gap " << s_string.length() - width << std::endl;
      }
      int gap = s_string.length() - width;
      if(gap > 0)
      {
	 if(gap < precision)
	 {
	    std::ostringstream another;
	    changeFormat(another, format);
	    another << std::setiosflags(std::ios::right | std::ios::adjustfield)
		    << std::setw(width) << std::setprecision(precision - gap) << xd;
	    s_string = another.str();
	 }
	 else
	 {
	    s_string += error_text;
	    throw StringException(factory_text + s_string);
	 }
      }
      return s_string;
   }

   template<class T>
   std::string StringFactory::intToString(const T xd, const int width)
   {
      std::string s_string;
      {
	 std::ostringstream s_stream;
	 s_stream << std::setiosflags(std::ios::right | std::ios::adjustfield)
		  << std::setw(width) << xd;
	 s_string = s_stream.str();
//       std::cerr << xd << ", width " << width
//            << ", gap " << s_string.length() - width << std::endl;
      }
      int gap = s_string.length() - width;
      if(gap > 0)
      {
	 s_string += error_text;
	 throw StringException(factory_text + s_string);
      }
      return s_string;
   }

#undef UTIL_USING_SSTREAM
#else

// deprecated strstream version
   template<class T>
   std::string StringFactory::floatToString(const T xd,
					    const int width, const int precision,
					    const char format)
   {
      std::string s_string;
      {
	 std::ostrstream s_stream;
	 s_stream.setf(std::ios::right, std::ios::adjustfield);
	 s_stream << std::setw(width) << std::setprecision(precision) << xd << std::ends;
	 s_string = s_stream.str();
	 s_stream.freeze(0);
//       std::cerr << xd << ", width " << width << ", prec " << precision 
//            << ", gap " << s_string.length() - width << std::endl;
      }
      int gap = s_string.length() - width;
      if(gap > 0)
      {
	 if(gap < precision)
	 {
	    std::ostrstream another;
	    another.setf(std::ios::right, std::ios::adjustfield);
	    another << std::setw(width) << std::setprecision(precision - gap) << xd << std::ends;
	    s_string = another.str();
	    another.freeze(0);
	 }
	 else
	 {
	    s_string += error_text;
	    throw StringException(factory_text + s_string);
	 }
      }
      return s_string;
   }

   template<class T>
   std::string StringFactory::intToString(const T xd, const int width)
   {
      std::string s_string;
      {
	 std::ostrstream s_stream;
	 s_stream.setf(std::ios::right, std::ios::adjustfield);
	 s_stream << std::setw(width) << xd << std::ends;
	 s_string = s_stream.str();
	 s_stream.freeze(0);
//       std::cerr << xd << ", width " << width
//            << ", gap " << s_string.length() - width << std::endl;
      }
      int gap = s_string.length() - width;
      if(gap > 0)
      {
	 s_string += error_text;
	 throw StringException(factory_text + s_string);
      }
      return s_string;
   }

#endif

}

#endif
