/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: testcomplex.cpp 600 2018-08-22 16:14:36Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C)  Niv Drory <drory@mpe.mpg.de>
 *                         Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * ---------------------------------------------------------------------
 *
 */

#define LTL_RANGE_CHECKING
//#define LTL_USE_SIMD

#include <ltl/marray.h>
#include <ltl/fvector.h>
#include <ltl/fmatrix.h>
#include <ltl/marray_io.h>
#include <ltl/statistics.h>
#include <iomanip>

using namespace ltl;

template<typename T>
void test_complex(void);
template<typename T>
void test_fv_complex(void);
template<typename T>
void test_fm_complex(void);

std::complex<float> user_func( std::complex<float> x )
{
   return x*x;
}
DECLARE_UNARY_FUNC( user_func, std::complex<float> )

#if defined(__GNUC__)
#  define THIS_GCC_VERSION (__GNUC__ * 10000 + __GNUC_MINOR__ * 100 + __GNUC_PATCHLEVEL__)
/* Test for GCC  3.6.x */
#  if (THIS_GCC_VERSION >= 40600 && THIS_GCC_VERSION < 40700)
#    define GCC_46_COMPLEX_DOUBLE_BUG
#  endif
#endif

int main(int argc, char **argv)
{
   std::cerr << "Testing MArray complex math ..." << std::endl;

   // this one test causes g++ to segfault on macosx intel with the apple-supplied g++=4.0 and 4.2
   // so we won't even try ...
#if defined(__APPLE__) && defined(__GNUC__) && !defined(__llvm__) && (__GNUC__==4 && __GNUC_MINOR__ < 3)
   cout << "Not testing test_complex<float> due to compiler bug." << std::endl;
#else
   test_complex<float>();
#endif
   test_complex<double>();

#ifdef GCC_46_COMPLEX_DOUBLE_BUG
   cout << "Not testing test_complex<double> due to gcc 4.6 compiler bug." << std::endl;
#endif

   std::cerr << "Testing FVector complex math ..." << std::endl;
   test_fv_complex<float>();
#ifndef GCC_46_COMPLEX_DOUBLE_BUG
   test_fv_complex<double>();
#endif

   std::cerr << "Testing FMatrix complex math ..." << std::endl;
   test_fm_complex<float>();
#ifndef GCC_46_COMPLEX_DOUBLE_BUG
   test_fm_complex<double>();
#endif

   std::cerr << "Testing user-suppplied function complex math ..." << std::endl;
   MArray<std::complex<float>,1> A(100), B(100);
   A = indexPosInt(A,1);
   B = user_func(A);
   LTL_ASSERT_( allof(B==A*A)," MArray user-supplied function on std::complex<T> failed" );
   return 0;
}


template<typename T>
void test_complex(void)
{
   MArray<std::complex<T>,1> A(10), B(10);
   A = std::complex<T>(1.0,2.0);
   std::complex<T> c(1.0,2.0);
   LTL_ASSERT_( allof(A==c)," MArray std::complex<T> assignment expression failed" );

   A = A + std::complex<T>(1.0);
   c = std::complex<T>(2.0,2.0);
   LTL_ASSERT_( allof(A==c)," MArray std::complex<T> simple add expression failed" );

   A = conj( A + std::complex<T>(1.0,-1.0) );
   c = std::complex<T>(3.0,-1.0);
   LTL_ASSERT_( allof(A==c)," MArray std::complex<T> conj() expression failed" );

   c = std::complex<T>(1.0,1.0);
   B = c;
   MArray<double,1> C;
   C = abs(B);
   LTL_ASSERT_( allof(C==abs(c))," MArray std::complex<T> abs() expression failed" );

   c = A(1);

   A = cos(  A + T(1.0) );
   A = exp(  A * A );
   A = log(  A + T(1.0) );
   A = sqrt( A + T(1.0) );
   A = tanh( A + T(1.0) );

   c = cos(  c + T(1.0) );
   c = exp(  c * c );
   c = log(  c + std::complex<T>(1.0) );
   c = sqrt( c + std::complex<T>(1.0) );
   c = tanh( c + std::complex<T>(1.0) );

   LTL_ASSERT_( allof(A==c)," MArray std::complex<T> transcendental expression failed" );
}

template<typename T>
void test_fv_complex(void)
{
   FVector<std::complex<T>,3> A, B;
   A = std::complex<T>(1.0,2.0);
   std::complex<T> c(1.0,2.0);
   LTL_ASSERT_( allof(A==c)," FVector std::complex<T> assignment expression failed" );

   A = A + std::complex<T>(1.0);
   c = c + std::complex<T>(1.0);
   LTL_ASSERT_( allof(A==c)," FVector std::complex<T> simple add expression failed" );

   A = conj( A + std::complex<T>(1.0,-1.0) );
   c = conj( c + std::complex<T>(1.0,-1.0) );
   LTL_ASSERT_( allof(A==c)," FVector std::complex<T> conj() expression failed" );

   c = std::complex<T>(1.0,1.0);
   B = c;
   FVector<T,3> C;
   C = abs(B);
   LTL_ASSERT_( allof(C==abs(c))," FVector std::complex<T> abs() expression failed" );

   c = A(1);

   A = cos(  A + T(1.0) );
   A = exp(  A * A );
   A = log(  A + T(1.0) );
   A = sqrt( A + T(1.0) );
   A = tanh( A + T(1.0) );

   c = cos(  c + T(1.0) );
   c = exp(  c * c );
   c = log(  c + T(1.0) );
   c = sqrt( c + T(1.0) );
   c = tanh( c + T(1.0) );

   LTL_ASSERT_( allof(A==c)," FVector std::complex<T> transcendental expression failed" );
}

template<typename T>
void test_fm_complex(void)
{
   FMatrix<std::complex<T>,3,4> A, B;
   A = std::complex<T>(1.0,2.0);
   std::complex<T> c(1.0,2.0);
   LTL_ASSERT_( allof(A==c)," FMatrix std::complex<T> assignment expression failed" );

   A = A + std::complex<T>(1.0);
   c = std::complex<T>(2.0,2.0);
   LTL_ASSERT_( allof(A==c)," FMatrix std::complex<T> simple add expression failed" );

   A = conj( A + std::complex<T>(1.0,-1.0) );
   c = std::complex<T>(3.0,-1.0);
   LTL_ASSERT_( allof(A==c)," FMatrix std::complex<T> conj() expression failed" );

   c = std::complex<T>(1.0,1.0);
   B = c;
   FMatrix<T,3,4> C;
   C = abs(B);
   LTL_ASSERT_( allof(C==abs(c))," FMatrix std::complex<T> abs() expression failed" );

   c = A(1,1);

   A = cos(  A + T(1.0) );
   A = exp(  A * A );
   A = log(  A + T(1.0) );
   A = sqrt( A + T(1.0) );
   A = tanh( A + T(1.0) );

   c = cos(  c + T(1.0) );
   c = exp(  c * c );
   c = log(  c + T(1.0) );
   c = sqrt( c + T(1.0) );
   c = tanh( c + T(1.0) );

   LTL_ASSERT_( allof(A==c)," FMatrix std::complex<T> transcendental expression failed" );
}
