/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: testfitsio.cpp 363 2008-07-10 15:09:44Z drory $
 * ---------------------------------------------------------------------
 *
 * Copyright (C)  Niv Drory <drory@mpe.mpg.de>
 *                         Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * ---------------------------------------------------------------------
 *
 */

//#define LTL_RANGE_CHECKING

#include <fitsio.h>
#include <statistics.h>
#include <marray_io.h>

#include <iostream>
#include <fstream>
#include <cstdio>

using namespace ltl;
using namespace util;

int main(int argc, char **argv);

void testextensionio( void )
{
   MArray<float,2> A(512,512), C;
   A = indexPosFlt(A,1) + 100.0*indexPosFlt(A,2);
   C = A+1.0f;
   
   // Write a FITS file with extensions
   
   FitsExtensionOut feo("testextensionio.fits");
   
   feo.addValueCard("TESTTEST", "This is a test");
   LTL_ASSERT_( feo.getString( "TESTTEST" ) == "This is a test",
                "FITS AddValueCard( string ) failed" );
   feo.addValueCard("TEST1234", 1.234);
   LTL_ASSERT_( feo.getFloat( "TEST1234" ) == 1.234,
                "FITS AddValueCard( float ) failed" ); 
   feo << emptyData; // primary with no data

   feo.addValueCard("EXTTEST ", "This is an extension test");
   LTL_ASSERT_( feo.getString( "EXTTEST " ) == "This is an extension test",
                "FITS AddValueCard( string ) failed" );
   feo << A;         // 1st extension with data

   feo << emptyData; // 2nd extension with no data

   feo.addValueCard("TEST5678", 5.678);
   LTL_ASSERT_( feo.getFloat( "TEST5678" ) == 5.678,
                "FITS AddValueCard( float ) failed" ); 
   feo.addValueCard("EXTNAME ", "This is an extension key test");
   LTL_ASSERT_( feo.getString( "EXTNAME " ) == "This is an extension key test",
                "FITS AddValueCard( string ) for extension key failed" );
   feo << C;  // 3rd extension with data

   // Now read that FITS file with extensions
   
   FitsExtensionIn fei("testextensionio.fits");

   LTL_ASSERT_( fei.getString( "TESTTEST" ) == "This is a test",
                "FITS getString() failed within primary header" );
   LTL_ASSERT_( fei.getFloat( "TEST1234" ) == 1.234,
                "FITS getFloat() failed within primary header" );
   try
   {      
      while(true)
      {
         FitsIn fi(fei.getNextExtension());

	 if( fei.getExtNo() == 3 ){
	    LTL_ASSERT_( fi.getFloat( "TEST5678" ) == 5.678,
			 "FITS AddValueCard( float ) extension test failed" ); 
	    LTL_ASSERT_( fi.getString( "EXTNAME " ) == "This is an extension key test",
			 "FITS AddValueCard( string ) for extension key failed" );
	 }
         else if( fei.getExtNo() % 2 )
         {         
            MArray<float,2> B;
            fi >> B;
            LTL_ASSERT_( allof(A == B),
                         "FITS extension data read or write failed");
            A += 1.0f;
         }
         LTL_ASSERT_( fi.getString( "TESTTEST" ) == "This is a test",
                      "FITS getString() failed within some extension header" );
         LTL_ASSERT_( fi.getFloat( "TEST1234" ) == 1.234,
                      "FITS getFloat() failed within some extension header" );
         LTL_ASSERT_( fi.getString( "EXTTEST " ) == "This is an extension test",
                      "FITS getString() test 2 failed within some extension header" );
      }
   }
   catch(FitsException e)
   {
      if(e.what() != std::string("no further extensions")) throw;
      //else std::cerr << "ending with: " << e.what() << std::endl;
   }

   remove("testextensionio.fits");
}

void testextensionioinmem( void )
{
   unsigned char* bufPtr = NULL;
   std::size_t bufSize = 0;
   std::ofstream fitsFile;

   MArray<float,2> A(512,512), C;
   A = indexPosFlt(A,1) + 100 * indexPosFlt(A,2);
   C = A + 1;
   
   // Write an in-memory FITS file with extensions
   
   FitsExtensionOut feo(&bufPtr, &bufSize);
   
   feo.addValueCard("EXTTYPE ", "core");
   LTL_ASSERT_( feo.getString( "EXTTYPE " ) == "core",
                "FITS AddValueCard( string ) failed" );

   feo << C; // primary with data
   //printf("after writing forst primary extension size is %d\n", (int)bufSize);
   
   // overwrite the first primary
   feo.resetExtNo();

   feo.addValueCard("EXTTYPE ", "raw");
   LTL_ASSERT_( feo.getString( "EXTTYPE " ) == "raw",
                "FITS AddValueCard( string ) failed" );

   feo << A; // primary with data
   //printf("after writing second primary extension size is %d\n", (int)bufSize);

   feo.addValueCard("EXTTYPE ", "clean");
   feo.addValueCard("EXTNAME ", "cleaned image");
   LTL_ASSERT_( feo.getString( "EXTNAME " ) == "cleaned image",
                "FITS AddValueCard( string ) failed" );

   // Uncomment these to force a getExtension error
   //feo.addValueCard("EXTNAME ", "cleaned image force error");
   //LTL_ASSERT_( feo.getString( "EXTNAME " ) == "cleaned image force error",
   //"FITS AddValueCard( string ) failed" );

   feo << C;         // 1st extension with data
   //printf("after writing 1st extension size is %d\n", (int)bufSize);

   fitsFile.exceptions(std::ofstream::failbit | std::ofstream::badbit);
   fitsFile.open("testextensionioinmem.fits", std::ofstream::out | std::ofstream::binary);
   fitsFile.write((char*)bufPtr, bufSize);
   fitsFile.close();

   // Now read back from the disk file.

   FitsIn*	pFi = NULL;

   {
      FitsExtensionIn fei("testextensionioinmem.fits");
      MArray<float,2> B;

      try
      {      
	 // Read extension 'cleaned image' using the disk file
	 // test extension by name
	 pFi = new FitsIn(fei.getExtension("IMAGE","cleaned image"));
      }
      catch(FitsException e)
      {
	 std::cerr << "Caught an exception: " << e.what() << std::endl;
	 if(e.what() != std::string("no further extensions"))
	    throw;
	 // Read the primary extension from the disk file
         pFi = &fei;
      }
	
      *pFi >> B;

      if (pFi == &fei)
      {
	 LTL_ASSERT_( allof(A == B),
		      "Data read mis-match");
	 LTL_ASSERT_( pFi->getString( "EXTTYPE " ) == "raw",
		      "FITS getString() failed within raw image header" );
      }
      else
      {
	 LTL_ASSERT_( allof(C == B),
		      "Data read mis-match");
	 LTL_ASSERT_( pFi->getString( "EXTTYPE " ) == "clean",
		      "FITS getString() failed within cleaned image extension header" );
	 delete pFi;
	 pFi = NULL;
      }
   }

   //std::cout << std::endl << "Now trying to read back from an in-memory file..." << std::endl;

   {
      FitsExtensionIn fei(bufPtr, bufSize);
      MArray<float,2> B;

      try
      {      
	 //std::cout << "Attempting to read extension 'cleaned image' using the in-memory file" << std::endl;
	 // test extension by name
	 pFi = new FitsIn(fei.getExtension("IMAGE","cleaned image"));
      }
      catch(FitsException e)
      {
	 std::cerr << "Caught an exception: " << e.what() << std::endl;
	 if(e.what() != std::string("no further extensions"))
	    throw;
	 //std::cout << "Reading the primary extension from the in-memory file" << std::endl;
         pFi = &fei;
      }
	
      *pFi >> B;
      //std::cout << "Data sucessfully read using the in-memory file" << std::endl;

      if (pFi == &fei)
      {
	 LTL_ASSERT_( allof(A == B),
		      "Data read mis-match");
	 LTL_ASSERT_( pFi->getString( "EXTTYPE " ) == "raw",
		      "FITS getString() failed within raw image header" );
      }
      else
      {
	 LTL_ASSERT_( allof(C == B),
		      "Data read mis-match");
	 LTL_ASSERT_( pFi->getString( "EXTTYPE " ) == "clean",
		      "FITS getString() failed within cleaned image extension header" );
	 delete pFi;
      }
   }

   remove("testextensionioinmem.fits");
}

int main(int argc, char **argv)
{
   const int offtsize = sizeof(off_t);
   if(sizeof(off_t)!=8){
      std::cerr << "Wrong offset type off_t: Must be 64 bit, but is "
		<< offtsize * 8 << " bit."<< std::endl;
      throw FitsException("");
   }

   std::cerr << "Testing MArray FITS extensions I/O ..." << std::endl;   
   try
   {
      testextensionio();
   }
   catch (std::exception &e){
      std::cerr << e.what() << std::endl;
      throw;
   }

   std::cerr << "Testing MArray FITS extensions in memory I/O ..." << std::endl;   
   try
   {
      testextensionioinmem();
   }
   catch (std::exception &e){
      std::cerr << e.what() << std::endl;
      throw;
   }
}
